@@ -1,3 +1,4 @@
+(define void (set! void 0))
 (define-syntax let
     (syntax-rules ()
     [(let ((name val) ...) body1 body2 ...)
@@ -35,8 +36,8 @@
     [(or) #f]
     [(or test) test]
     [(or test1 test2 ...)
-     (let ((x test1))
-       (if x x (or test2 ...)))]))
+     (let ((var1 test1))
+       (if var1 var1 (or test2 ...)))]))
 
 (define-syntax and
  (syntax-rules ()
@@ -98,6 +99,39 @@
              (begin result1 result2 ...)
              (cond clause1 clause2 ...)))))
 
+(define-syntax case
+  (syntax-rules (else =>)
+    ((case (key ...)
+       clauses ...)
+     (let ((atom-key (key ...)))
+       (case atom-key clauses ...)))
+    ((case key
+       (else => result))
+     (result key))
+    ((case key
+       (else result1 result2 ...))
+     (begin result1 result2 ...))
+    ((case key
+       ((atoms ...) result1 result2 ...))
+     (if (memv key '(atoms ...))
+         (begin result1 result2 ...)))
+    ((case key
+       ((atoms ...) => result))
+     (if (memv key '(atoms ...))
+         (result key)))
+    ((case key
+       ((atoms ...) => result)
+       clause clauses ...)
+     (if (memv key '(atoms ...))
+         (result key)
+         (case key clause clauses ...)))
+    ((case key
+       ((atoms ...) result1 result2 ...)
+       clause clauses ...)
+     (if (memv key '(atoms ...))
+         (begin result1 result2 ...)
+         (case key clause clauses ...)))))
+
 (define (atom? obj)
     (and (not (pair? obj)) (not (vector? obj))))
 
@@ -153,4 +187,72 @@
     (else (assoc obj (cdr alist)))))
 
 (define (substring string start end)
-    (string-copy string start end))
\ No newline at end of file
+    (string-copy string start end))
+
+(define-syntax delay-force
+  (syntax-rules ()
+    ((delay-force expression)
+     (make-promise #f (lambda () expression)))))
+
+(define-syntax delay
+   (syntax-rules ()
+     ((delay expression)
+      (delay-force (make-promise #t expression)))))
+
+(define make-promise
+    (lambda (done? proc)
+      (list (cons done? proc))))
+
+(define (force promise)
+    (if (promise-done? promise)
+        (promise-value promise)
+        (let ((promise* ((promise-value promise))))
+          (unless (promise-done? promise)
+            (promise-update! promise* promise))
+          (force promise))))
+
+(define promise-done?
+    (lambda (x) (car (car x))))
+  (define promise-value
+    (lambda (x) (cdr (car x))))
+  (define promise-update!
+    (lambda (new old)
+      (set-car! (car old) (promise-done? new))
+      (set-cdr! (car old) (promise-value new))
+      (set-car! new (car old))))
+
+(define (add1 x) (+ x 1))
+(define (sub1 x) (- x 1))
+(define (newline) (display #\newline))
+
+(define (any? proc list)
+    (and (pair? list)
+        (or (proc (car list))
+             (any? proc (cdr list)))))
+
+(define (map1 f xs)
+  (if (null? xs)
+      '()
+      (cons (f (car xs)) (map1 f (cdr xs)))))
+
+(define (map f . xss)
+  (letrec
+   ((map-all
+     (lambda (xss)
+       (if (any? null? xss)
+           '()
+           (cons (apply f (map1 car xss))
+                 (map-all (map1 cdr xss)))))))
+
+    (map-all xss)))
+
+(define (for-each f . xss)
+  (letrec
+   ((for-each-all
+     (lambda (xss)
+       (if (any? null? xss)
+           void
+           (begin (apply f (map1 car xss))
+                  (for-each-all (map1 cdr xss)) void)))))
+
+    (for-each-all xss)))
