@@ -20,9 +20,8 @@ pub enum Cell {
     // Types that exist in VCell, but need Cell representation for
     // printing purposes. These are never created by the lexer/parser.
     Continuation,
-    Closure,
     Macro,
-    Lambda,
+    Procedure(Option<String>),
     Undefined,
     Void,
 }
@@ -372,16 +371,28 @@ impl Display for Cell {
                 loop {
                     match (*cdr).as_ref() {
                         Cell::Nil => {
-                            write!(f, "{:#})", car)?;
+                            if f.alternate() {
+                                write!(f, "{:#})", car)?;
+                            } else {
+                                write!(f, "{})", car)?;
+                            }
                             return Ok(());
                         }
                         Cell::Pair(ncar, ncdr) => {
-                            write!(f, "{:#} ", car)?;
+                            if f.alternate() {
+                                write!(f, "{:#} ", car)?;
+                            } else {
+                                write!(f, "{} ", car)?;
+                            }
                             car = ncar;
                             cdr = ncdr;
                         }
                         _ => {
-                            write!(f, "{:#} . {:#})", car, cdr)?;
+                            if f.alternate() {
+                                write!(f, "{:#} . {:#})", car, cdr)?;
+                            } else {
+                                write!(f, "{} . {})", car, cdr)?;
+                            }
                             return Ok(());
                         }
                     }
@@ -407,8 +418,11 @@ impl Display for Cell {
                             '\t' => write!(f, "\\t")?,
                             '\n' => write!(f, "\\n")?,
                             '\r' => write!(f, "\\r")?,
+                            _ if it as u32 == 0x1b => write!(f, "\\e")?,
                             _ if it as u32 == 0x7 => write!(f, "\\a")?,
                             _ if it as u32 == 0x8 => write!(f, "\\b")?,
+                            _ if it as u32 == 0xb => write!(f, "\\v")?,
+                            _ if it as u32 == 0xc => write!(f, "\\f")?,
                             _ if it.is_ascii_control() => write!(f, "\\x{:x};", it as u32)?,
                             it => write!(f, "{}", it)?,
                         };
@@ -426,9 +440,15 @@ impl Display for Cell {
                 write!(f, "#(")?;
                 for (idx, cell) in vector.iter().enumerate() {
                     if idx == vector.len() - 1 {
-                        write!(f, "{:#}", cell)?;
-                    } else {
+                        if f.alternate() {
+                            write!(f, "{:#}", cell)?;
+                        } else {
+                            write!(f, "{}", cell)?;
+                        }
+                    } else if f.alternate() {
                         write!(f, "{:#} ", cell)?;
+                    } else {
+                        write!(f, "{} ", cell)?;
                     }
                 }
                 write!(f, ")")
@@ -436,15 +456,17 @@ impl Display for Cell {
             Cell::Continuation => {
                 write!(f, "#<continuation>")
             }
-            Cell::Closure => {
-                write!(f, "#<procedure>")
-            }
             Cell::Macro => {
                 write!(f, "#<macro>")
             }
-            Cell::Lambda => {
-                write!(f, "#<procedure>")
-            }
+            Cell::Procedure(desc) => match desc {
+                Some(desc) => {
+                    write!(f, "#<procedure:{}>", desc)
+                }
+                None => {
+                    write!(f, "#<procedure>")
+                }
+            },
             Cell::Undefined => {
                 write!(f, "#<undefined>")
             }
