@@ -1,5 +1,8 @@
 use num::bigint::BigInt;
-use num::{BigRational, CheckedAdd, CheckedDiv, CheckedMul, CheckedSub, FromPrimitive, Signed};
+use num::traits::FloatConst;
+use num::{
+    BigRational, CheckedAdd, CheckedDiv, CheckedMul, CheckedSub, FromPrimitive, Rational64, Signed,
+};
 use num::{Num, Rational32, ToPrimitive};
 use std::cmp::Ordering;
 use std::fmt;
@@ -40,7 +43,7 @@ pub enum Exactness {
 /// * Exactness if preferred
 /// * Inexactness is infectious
 ///
-/// Rationsls are provided by the Rational32 type, allowing
+/// Rationals are provided by the Rational32 type, allowing
 /// for a rational composed of a 32 bit numerator and denominator.
 /// If a rational exceeds these limits, the resulting operation
 /// falls back to "inexact" (float).
@@ -120,21 +123,23 @@ impl Number {
 
     pub fn to_usize(&self) -> Option<usize> {
         match self {
-            Number::Fixnum(num) if *num >= 0 => Some(*num as usize),
+            Number::Fixnum(num) if *num >= 0 => num.to_usize(),
             Number::BigInt(num)
                 if **num >= BigInt::from(0) && **num <= BigInt::from(usize::MAX) =>
             {
                 Some(num.to_usize().unwrap())
             }
+            Number::Rational(num) if num.is_integer() => num.to_usize(),
             _ => None,
         }
     }
 
     pub fn to_i64(&self) -> Option<i64> {
         match self {
-            Number::Fixnum(num) => Some(*num as i64),
+            Number::Fixnum(num) => Some(*num),
             Number::BigInt(num) => num.to_i64(),
             Number::Rational(num) if num.is_integer() => num.to_i64(),
+            Number::Float(num) if self.is_integer() => num.to_i64(),
             _ => None,
         }
     }
@@ -146,25 +151,36 @@ impl Number {
                 Some(num.to_u64().unwrap())
             }
             Number::Rational(num) if num.is_integer() => num.to_u64(),
+            Number::Float(num) if self.is_integer() => num.to_u64(),
             _ => None,
         }
     }
 
     pub fn to_u32(&self) -> Option<u32> {
         match self {
-            Number::Fixnum(num) if *num >= 0 => Some(*num as u32),
+            Number::Fixnum(num) => num.to_u32(),
             Number::BigInt(num) if **num >= BigInt::from(0) && **num <= BigInt::from(u32::MAX) => {
                 Some(num.to_u32().unwrap())
             }
             Number::Rational(num) if num.is_integer() => num.to_u32(),
+            Number::Float(num) if self.is_integer() => num.to_u32(),
             _ => None,
         }
     }
 
+    pub fn to_f64(&self) -> Option<f64> {
+        match self {
+            Number::Fixnum(num) => num.to_f64(),
+            Number::BigInt(num) => num.to_f64(),
+            Number::Rational(num) => num.to_f64(),
+            Number::Float(num) => Some(*num),
+        }
+    }
+
     pub fn is_integer(&self) -> bool {
         match self {
             Number::Fixnum(_) => true,
-            Number::Float(_) => false,
+            Number::Float(num) => num.floor() == *num,
             Number::BigInt(_) => true,
             Number::Rational(num) => num.is_integer(),
         }
@@ -202,10 +218,19 @@ impl Number {
 
     pub fn to_exact(&self) -> Option<Number> {
         match self {
-            Number::Float(num) => match Rational32::from_f64(*num) {
-                Some(num) => Some(num.into()),
-                None => Some((*num).into()),
-            },
+            Number::Float(num) => {
+                if self.is_integer() {
+                    match num.to_i64() {
+                        Some(integer) => Some(Number::Fixnum(integer)),
+                        None => num.to_i128().map(|num| BigInt::from(num).into()),
+                    }
+                } else {
+                    match Rational32::from_f64(*num) {
+                        Some(num) => Some(num.into()),
+                        None => Some((*num).into()),
+                    }
+                }
+            }
             Number::BigInt(_) | Number::Rational(_) | Number::Fixnum(_) => Some(self.clone()),
         }
     }
@@ -236,7 +261,7 @@ impl Number {
 
     pub fn abs(&self) -> Number {
         match self {
-            Number::Fixnum(num) => num.abs().into(),
+            Number::Fixnum(num) => num.unsigned_abs().into(),
             Number::Float(num) => num.abs().into(),
             Number::BigInt(num) => num.abs().into(),
             Number::Rational(num) => num.abs().into(),
@@ -775,12 +800,17 @@ impl Div for &Number {
 }
 
 impl Number {
+    /// Quotient with truncation (rounding towards zero).
+    ///
+    /// The spec only defines this function for integer inputs. Since that includes integral floats,
+    /// we define quotient over all floats with the expectation that the caller can check the
+    /// inputs for strict conformance.
     pub fn quotient(&self, rhs: &Self) -> Option<Number> {
         match self {
             Number::Fixnum(lhs) => match rhs {
                 Number::Fixnum(rhs) => Some((lhs / rhs).into()),
                 Number::BigInt(rhs) => Some((BigInt::from(*lhs) / &**rhs).into()),
-                Number::Float(_) => None,
+                Number::Float(rhs) => lhs.to_f64().map(|lhs| (lhs as f64 / rhs).trunc().into()),
                 Number::Rational(rhs) => {
                     if rhs.is_integer() {
                         Some((*lhs / rhs.to_i64().unwrap()).into())
@@ -792,7 +822,7 @@ impl Number {
             Number::BigInt(lhs) => match rhs {
                 Number::Fixnum(rhs) => Some((&**lhs / rhs).into()),
                 Number::BigInt(rhs) => Some((&**lhs / &**rhs).into()),
-                Number::Float(_) => None,
+                Number::Float(rhs) => lhs.to_f64().map(|lhs| (lhs / rhs).trunc().into()),
                 Number::Rational(rhs) => {
                     if rhs.is_integer() {
                         Some((&**lhs / BigInt::from(rhs.to_i32().unwrap())).into())
@@ -801,15 +831,15 @@ impl Number {
                     }
                 }
             },
-            Number::Float(_) => match rhs {
-                Number::Fixnum(_) => None,
-                Number::Float(_) => None,
+            Number::Float(lhs) => match rhs {
+                Number::Fixnum(rhs) => Some((lhs / *rhs as f64).into()),
+                Number::Float(rhs) => Some((lhs / rhs).trunc().into()),
                 Number::BigInt(_) => None,
-                Number::Rational(_) => None,
+                Number::Rational(rhs) => rhs.to_f64().map(|rhs| (lhs / rhs).into()),
             },
             Number::Rational(lhs) if lhs.is_integer() => match rhs {
                 Number::Fixnum(rhs) => Some((lhs.to_i64().unwrap() / *rhs).into()),
-                Number::Float(_) => None,
+                Number::Float(rhs) => lhs.to_f64().map(|lhs| (lhs / rhs).trunc().into()),
                 Number::BigInt(rhs) => Some((BigInt::from(lhs.to_i64().unwrap()) / &**rhs).into()),
                 Number::Rational(rhs) => {
                     if rhs.is_integer() {
@@ -834,18 +864,22 @@ impl Rem for Number {
 impl Rem for &Number {
     type Output = Option<Number>;
 
+    /// Remainder
+    ///
+    /// The spec only defines remainder for integers but this operation is also used by other
+    /// functions that deal with numbers of all types internally.
     fn rem(self, rhs: Self) -> Self::Output {
         match self {
             Number::Fixnum(lhs) => match rhs {
                 Number::Fixnum(rhs) => Some((lhs % rhs).into()),
                 Number::BigInt(rhs) => Some((BigInt::from(*lhs) % &**rhs).into()),
-                Number::Float(_) => None,
+                Number::Float(rhs) => Some((*lhs as f64 % rhs).into()),
                 Number::Rational(rhs) => {
-                    if rhs.is_integer() {
-                        Some((*lhs % rhs.to_i64().unwrap()).into())
-                    } else {
-                        None
-                    }
+                    let result = Rational64::from_integer(*lhs)
+                        % Rational64::from((*rhs.numer() as i64, *rhs.denom() as i64));
+                    // lhs % rhs is guaranteed to be less than rhs. Since lhs is a whole integer,
+                    // both numerator and denominator still fit in an i32.
+                    Some(Rational32::from((*result.numer() as i32, *result.denom() as i32)).into())
                 }
             },
             Number::BigInt(lhs) => match rhs {
@@ -854,31 +888,33 @@ impl Rem for &Number {
                 Number::Float(_) => None,
                 Number::Rational(rhs) => {
                     if rhs.is_integer() {
+                        // fast path if rhs is an integer
                         Some((&**lhs % BigInt::from(rhs.to_i32().unwrap())).into())
                     } else {
-                        None
+                        // if rhs is a rational, we can mod the bigint by the numerator
+                        // and then divide the result by the denominator.
+                        // Exercise to the reader: Prove that X % N/M == (X % N) / M
+                        let denom = rhs.denom();
+                        let numer_mod =
+                            (&**lhs * BigInt::from(*denom)) % BigInt::from(*rhs.numer());
+                        Some(Rational32::new(numer_mod.to_i32().unwrap(), *denom).into())
                     }
                 }
             },
-            Number::Float(_) => match rhs {
-                Number::Fixnum(_) => None,
-                Number::Float(_) => None,
-                Number::BigInt(_) => None,
-                Number::Rational(_) => None,
+            Number::Float(lhs) => match rhs {
+                Number::Fixnum(rhs) => Some((lhs % *rhs as f64).into()),
+                Number::Float(rhs) => Some((lhs % rhs).into()),
+                // BigInt::to_f64 converts to infinity if it is too large
+                // which yields the correct result since a % b = a if a < b
+                Number::BigInt(rhs) => rhs.to_f64().map(|rhs| (lhs % rhs).into()),
+                Number::Rational(rhs) => rhs.to_f64().map(|rhs| (lhs % rhs).into()),
             },
-            Number::Rational(lhs) if lhs.is_integer() => match rhs {
+            Number::Rational(lhs) => match rhs {
                 Number::Fixnum(rhs) => Some((lhs.to_i64().unwrap() % *rhs).into()),
-                Number::Float(_) => None,
+                Number::Float(rhs) => lhs.to_f64().map(|lhs| (lhs % rhs).into()),
                 Number::BigInt(rhs) => Some((BigInt::from(lhs.to_i64().unwrap()) % &**rhs).into()),
-                Number::Rational(rhs) => {
-                    if rhs.is_integer() {
-                        Some((lhs % rhs).into())
-                    } else {
-                        None
-                    }
-                }
+                Number::Rational(rhs) => Some((lhs % rhs).into()),
             },
-            Number::Rational(_) => None,
         }
     }
 }
@@ -900,6 +936,7 @@ impl fmt::Display for Number {
             Number::Fixnum(num) => write!(f, "{}", num),
             Number::BigInt(num) => write!(f, "{}", num),
             Number::Float(num) if *num > 1E10 => write!(f, "{:e}", num),
+            Number::Float(num) if self.is_integer() => write!(f, "{:.1}", num),
             Number::Float(num) => write!(f, "{}", num),
             Number::Rational(num) => write!(f, "{}", num),
         }
@@ -996,6 +1033,16 @@ impl From<u64> for Number {
     }
 }
 
+impl From<usize> for Number {
+    fn from(num: usize) -> Self {
+        if num > i64::MAX as usize {
+            Number::new_bigint(BigInt::from(num))
+        } else {
+            Number::Fixnum(num as i64)
+        }
+    }
+}
+
 impl From<i64> for Number {
     fn from(num: i64) -> Self {
         Number::Fixnum(num)
@@ -1020,11 +1067,121 @@ impl From<Rational32> for Number {
     }
 }
 
+/// Trigonometric functions
+impl Number {
+    pub fn exp(&self) -> Option<Number> {
+        match self {
+            Number::Fixnum(num) => num.to_f64().map(|num| num.exp().into()),
+            Number::Float(num) => Some(num.exp().into()),
+            Number::BigInt(num) => num.to_f64().map(|num| num.exp().into()),
+            Number::Rational(num) => num.to_f64().map(|num| num.exp().into()),
+        }
+    }
+
+    pub fn log(&self) -> Option<Number> {
+        match self {
+            Number::Fixnum(num) => num.to_f64().map(|num| num.log(f64::E()).into()),
+            Number::Float(num) => Some(num.log(f64::E()).into()),
+            Number::BigInt(num) => num.to_f64().map(|num| num.log(f64::E()).into()),
+            Number::Rational(num) => num.to_f64().map(|num| num.log(f64::E()).into()),
+        }
+    }
+
+    pub fn sin(&self) -> Option<Number> {
+        match self {
+            Number::Fixnum(num) => num.to_f64().map(|num| num.sin().into()),
+            Number::Float(num) => Some(num.sin().into()),
+            Number::BigInt(num) => num.to_f64().map(|num| num.sin().into()),
+            Number::Rational(num) => num.to_f64().map(|num| num.sin().into()),
+        }
+    }
+
+    pub fn cos(&self) -> Option<Number> {
+        match self {
+            Number::Fixnum(num) => num.to_f64().map(|num| num.cos().into()),
+            Number::Float(num) => Some(num.cos().into()),
+            Number::BigInt(num) => num.to_f64().map(|num| num.cos().into()),
+            Number::Rational(num) => num.to_f64().map(|num| num.cos().into()),
+        }
+    }
+
+    pub fn tan(&self) -> Option<Number> {
+        match self {
+            Number::Fixnum(num) => num.to_f64().map(|num| num.tan().into()),
+            Number::Float(num) => Some(num.tan().into()),
+            Number::BigInt(num) => num.to_f64().map(|num| num.tan().into()),
+            Number::Rational(num) => num.to_f64().map(|num| num.tan().into()),
+        }
+    }
+
+    pub fn asin(&self) -> Option<Number> {
+        match self {
+            Number::Fixnum(num) => num.to_f64().map(|num| num.asin().into()),
+            Number::Float(num) => Some(num.asin().into()),
+            Number::BigInt(num) => num.to_f64().map(|num| num.asin().into()),
+            Number::Rational(num) => num.to_f64().map(|num| num.asin().into()),
+        }
+    }
+
+    pub fn acos(&self) -> Option<Number> {
+        match self {
+            Number::Fixnum(num) => num.to_f64().map(|num| num.acos().into()),
+            Number::Float(num) => Some(num.acos().into()),
+            Number::BigInt(num) => num.to_f64().map(|num| num.acos().into()),
+            Number::Rational(num) => num.to_f64().map(|num| num.acos().into()),
+        }
+    }
+
+    pub fn atan(&self) -> Option<Number> {
+        match self {
+            Number::Fixnum(num) => num.to_f64().map(|num| num.atan().into()),
+            Number::Float(num) => Some(num.atan().into()),
+            Number::BigInt(num) => num.to_f64().map(|num| num.atan().into()),
+            Number::Rational(num) => num.to_f64().map(|num| num.atan().into()),
+        }
+    }
+
+    pub fn atan2(&self, x: Number) -> Option<Number> {
+        if let Some(x) = x.to_f64() {
+            match self {
+                Number::Fixnum(num) => num.to_f64().map(|num| num.atan2(x).into()),
+                Number::Float(num) => Some(num.atan2(x).into()),
+                Number::BigInt(num) => num.to_f64().map(|num| num.atan2(x).into()),
+                Number::Rational(num) => num.to_f64().map(|num| num.atan2(x).into()),
+            }
+        } else {
+            None
+        }
+    }
+
+    /// sqrt is undefined for complex results
+    pub fn sqrt(&self) -> Option<Number> {
+        let result: Option<Number> = match self {
+            Number::Fixnum(num) => num.to_f64().map(|num| num.sqrt().into()),
+            Number::Float(num) => Some(num.sqrt().into()),
+            Number::BigInt(num) => num.to_f64().map(|num| num.sqrt().into()),
+            Number::Rational(num) => num.to_f64().map(|num| num.sqrt().into()),
+        };
+
+        // convert NaN and infinity to None
+        if let Some(r) = &result {
+            if r.to_f64().unwrap().is_finite() {
+                result
+            } else {
+                None
+            }
+        } else {
+            result
+        }
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use crate::number::Number;
     use num::bigint::BigInt;
-    use num::Rational32;
+    use num::traits::FloatConst;
+    use num::{FromPrimitive, Rational32};
     use std::mem;
     use std::str::FromStr;
 
@@ -1047,6 +1204,7 @@ mod tests {
         assert!(Number::from(1.0).to_usize().is_none());
         let overflow_num = Number::from(BigInt::from(usize::MAX)) + Number::from(1);
         assert!(overflow_num.to_usize().is_none());
+        assert_eq!(Number::from(Rational32::new(10, 1)).to_usize(), Some(10));
     }
 
     #[test]
@@ -1248,8 +1406,10 @@ mod tests {
             i32_oflow, BigInt::from(i32_oflow) => 1.0,
             i32_oflow, BigInt::from(2) => i32_oflow as f64 / 2_f64,
             100, Rational32::from_integer(50) => Rational32::from_integer(2),
+            100, Rational32::new(1, 2) => Rational32::from_integer(200),
             i32_oflow, Rational32::from_integer(2) => i32_oflow as f64 / 2_f64,
-            i32::MAX as i64, Rational32::new(1, 2) => i32::MAX as f64 * 2.0
+            i32::MAX as i64, Rational32::new(1, 2) => i32::MAX as f64 * 2.0,
+            -100, 1.0 => -100.0
         ];
 
         // BIGINT / RHS
@@ -1299,14 +1459,25 @@ mod tests {
             Number::from(100) % Number::from(BigInt::from(i64::MAX) + BigInt::from(i64::MAX)),
             Some(Number::from(BigInt::from(100)))
         );
-        assert_eq!(Number::from(100) % Number::from(1.0), None);
+        assert_eq!(
+            Number::from(100) % Number::from(1.0),
+            Some(Number::Float(0.0))
+        );
         assert_eq!(
             Number::from(100) % Number::from(Rational32::from_integer(70)),
             Some(Number::from(30))
         );
         assert_eq!(
             Number::from(100) % Number::from(Rational32::new(1, 2)),
-            None
+            Some(Number::from(0))
+        );
+        assert_eq!(
+            Number::from(1) % Number::from(Rational32::new(2, 5)),
+            Some(Number::from(Rational32::new(1, 5)))
+        );
+        assert_eq!(
+            Number::from(1) % Number::from(Rational32::new(3, 5)),
+            Some(Number::from(Rational32::new(2, 5)))
         );
 
         // BIGINT % RHS
@@ -1318,14 +1489,15 @@ mod tests {
             Number::from(BigInt::from(100)) % Number::from(BigInt::from(70)),
             Some(Number::from(BigInt::from(30)))
         );
+        // cannot support bigint % float (yet?)
         assert_eq!(Number::from(BigInt::from(100)) % Number::from(1.0), None);
         assert_eq!(
             Number::from(BigInt::from(100)) % Number::from(Rational32::from_integer(70)),
             Some(Number::from(BigInt::from(30)))
         );
         assert_eq!(
-            Number::from(BigInt::from(100)) % Number::from(Rational32::new(1, 2)),
-            None
+            Number::from(BigInt::from(1)) % Number::from(Rational32::new(2, 3)),
+            Some(Number::from(Rational32::new(1, 3)))
         );
 
         // RATIONAL % RHS
@@ -1342,8 +1514,49 @@ mod tests {
                 % Number::from(Rational32::from_integer(70)),
             Some(Number::from(Rational32::from_integer(30)))
         );
+        assert_eq!(
+            Number::from(Rational32::from_integer(1)) % Number::from(Rational32::new(2, 3)),
+            Some(Number::from(Rational32::new(1, 3)))
+        );
 
-        assert_eq!(Number::from(Rational32::new(1, 2)) % Number::from(70), None);
+        // FLOAT % RHS
+        assert_eq!(
+            Number::from(0.5) % Number::from(70),
+            Some(Number::from(0.5))
+        );
+        assert_eq!(
+            Number::from(100.0) % Number::from(70),
+            Some(Number::from(30.0))
+        );
+        assert!((Number::from(f64::INFINITY) % Number::from(70))
+            .unwrap()
+            .to_f64()
+            .unwrap()
+            .is_nan());
+        assert_eq!(
+            Number::from(100.0) % Number::from(BigInt::from(70)),
+            Some(Number::from(30.0))
+        );
+        assert_eq!(
+            Number::from(100.0) % Number::from(BigInt::from(i64::MAX) + BigInt::from(i64::MAX)),
+            Some(Number::from(100.0))
+        );
+        assert_eq!(
+            Number::from(101.0) % Number::from(2.0),
+            Some(Number::from(1.0))
+        );
+        assert_eq!(
+            Number::from(100.0) % Number::from(Rational32::from_integer(70)),
+            Some(Number::from(30.0))
+        );
+        assert_eq!(
+            Number::from(100.0) % Number::from(Rational32::new(1, 2)),
+            Some(Number::from(0.0))
+        );
+        assert_eq!(
+            Number::from(0.75) % Number::from(Rational32::new(1, 2)),
+            Some(Number::from(0.25))
+        );
     }
 
     #[test]
@@ -1368,4 +1581,33 @@ mod tests {
 
         assert_eq!(Number::from(Rational32::new(7, 2)).round(), Number::from(4));
     }
+
+    #[test]
+    fn abs() {
+        assert_eq!(Number::from(-10).abs(), Number::from(10));
+        assert_eq!(
+            Number::from(i64::MIN).abs(),
+            Number::from(BigInt::from(9223372036854775808_u64))
+        );
+    }
+
+    #[test]
+    fn sin() {
+        assert_eq!(Number::from(0).sin(), Some(Number::from(0)));
+        assert_eq!(
+            Number::from(f64::FRAC_PI_2()).sin(),
+            Some(Number::from(1.0))
+        );
+        assert_eq!(
+            Number::from(Rational32::from_integer(0)).sin(),
+            Some(Number::from(0.0))
+        );
+        assert_eq!(Number::from(BigInt::from(0)).sin(), Some(Number::from(0.0)));
+        assert!(Number::from(BigInt::from_f64(f64::MAX).unwrap() * 2)
+            .sin()
+            .unwrap()
+            .to_f64()
+            .unwrap()
+            .is_nan());
+    }
 }
