@@ -31,18 +31,18 @@ macro_rules! cdr {
 }
 
 impl Vm {
-    /// Compile
+    /// Compile Runnable
     ///
     /// Compile compiles a single top level expression, returning a Lambda that represents
     /// the top level expression of an Error if compilation failed.
     ///
     /// `expr` - The expression to compile.
-    pub fn compile(&mut self, expr: &Cell) -> Result<Lambda, Error> {
+    pub fn compile_runnable(&mut self, expr: &Cell) -> Result<Lambda, Error> {
         let mut entry_lambda = Lambda::new(vec![]);
         let mut lambda = Lambda::new_from_iof(vec![], vec![], &entry_lambda, &[], false);
         lambda.set_top_level();
         lambda.emit(OpCode::Enter);
-        self.compile_expression(&mut lambda, true, expr)?;
+        self.compile(&mut lambda, true, expr)?;
         lambda.emit(OpCode::Ret);
         trace!("main: \n{}", self.decompile_text(&lambda));
         let lambda = self.heap.put(lambda);
@@ -56,6 +56,67 @@ impl Vm {
         Ok(entry_lambda)
     }
 
+    /// Compile
+    ///
+    /// Apply any pre-compilation transforms to expr, compile it, and return
+    /// the resulting bytecode.
+    ///
+    /// # Arguments
+    /// `expr` - The expression to compile.
+    pub fn compile(&mut self, lambda: &mut Lambda, tail: bool, expr: &Cell) -> Result<(), Error> {
+        trace!("transforming {}", expr);
+        let expr = self.transform(expr)?;
+        trace!("compiling {}", expr);
+        self.compile_expression(lambda, tail, &expr)?;
+        Ok(())
+    }
+
+    /// Transform
+    ///
+    /// Apply pre-compilation transforms to expr, returning the transformed
+    /// AST.
+    pub fn transform(&mut self, expr: &Cell) -> Result<Cell, Error> {
+        match expr {
+            Cell::Pair(_, _) => self.transform_procedure_application(expr),
+            cell => Ok(cell.clone()),
+        }
+    }
+
+    pub fn transform_procedure_application(&mut self, expr: &Cell) -> Result<Cell, Error> {
+        let proc = expr.car().unwrap();
+        let mut rest = expr.cdr().unwrap();
+
+        if let Cell::Symbol(proc) = proc.deref() {
+            if let "quote" | "define-syntax" = proc.as_str() {
+                return Ok(expr.clone());
+            }
+        }
+
+        if let Some(sym) = self.heap.get_sym_ref(proc) {
+            let vcell = match self.globenv.get(sym.as_ptr()?) {
+                Some(VCell::Ptr(ptr)) => Some(self.heap.get_at_index(ptr).clone()),
+                vcell => vcell,
+            };
+            if let Some(VCell::Macro(transform)) = vcell {
+                let expansion = transform.transform(expr)?;
+                trace!("macro expansion: {} => {}", expr, expansion);
+                return self.transform(&expansion);
+            }
+        }
+
+        let mut v = vec![self.transform(proc)?];
+        while rest.is_pair() {
+            v.push(self.transform(rest.car().unwrap())?);
+            rest = rest.cdr().unwrap();
+        }
+        if rest.is_nil() {
+            Ok(Cell::new_list(v))
+        } else {
+            let rest = self.transform(rest)?;
+            Ok(Cell::new_improper_list(v, rest))
+        }
+    }
+
     /// Compile Expression
     ///
     /// Compile expression compiles a single expression, emitting its byte code to the currently
@@ -75,17 +136,14 @@ impl Vm {
             Cell::Pair(_, _) => self.compile_procedure_application(lambda, tail, expr),
             Cell::Symbol(_) => self.compile_symbol_expression(lambda, expr),
             Cell::Nil => Err(UnquotedNil),
-            Cell::Bool(_)
-            | Cell::Continuation
-            | Cell::Char(_)
-            | Cell::Number(_)
-            | Cell::Closure
-            | Cell::Macro
-            | Cell::String(_)
-            | Cell::Undefined
-            | Cell::Vector(_)
+            Cell::Procedure(_)
             | Cell::Void
-            | Cell::Lambda => self.compile_quote(lambda, expr),
+            | Cell::Undefined
+            | Cell::Macro
+            | Cell::Continuation => Err(InvalidSyntax(expr.to_string())),
+            Cell::Bool(_) | Cell::Char(_) | Cell::Number(_) | Cell::String(_) | Cell::Vector(_) => {
+                self.compile_quote(lambda, expr)
+            }
         }
     }
 
@@ -110,21 +168,6 @@ impl Vm {
     ) -> Result<(), Error> {
         let proc = expr.car().unwrap();
         let rest = expr.cdr().unwrap();
-
-        if let Some(sym) = self.heap.get_sym_ref(proc) {
-            let vcell = match self.globenv.get(sym.as_ptr()?) {
-                Some(VCell::Ptr(ptr)) => Some(self.heap.get_at_index(ptr).clone()),
-                vcell => vcell,
-            };
-
-            if let Some(VCell::Macro(transform)) = vcell {
-                let expansion = transform.transform(expr)?;
-                trace!("macro expansion: {} => {}", expr, expansion);
-                self.compile_expression(lambda, tail, &expansion)?;
-                return Ok(());
-            }
-        }
-
         match proc.deref() {
             Cell::Symbol(proc) => match proc.as_str() {
                 "define" => self.compile_define(lambda, expr),
@@ -242,7 +285,7 @@ impl Vm {
         }
 
         lambda.emit(OpCode::MovImmediate);
-        lambda.emit(VCell::void());
+        lambda.emit(self.heap.put(VCell::Void));
         lambda.emit(VCell::Acc);
         Ok(())
     }
@@ -364,13 +407,13 @@ impl Vm {
 
         // The position of formal args and body differ slightly
         // on whether this was a define special form or a lambda
-        let (formal_args, body) = match is_define_special {
+        let (formal_args_ast, body) = match is_define_special {
             true => (cdr!(car!(rest)), cdr!(rest)),
             false => (car!(rest), cdr!(rest)),
         };
 
         // Compile formal args into a list of symbols
-        let (formal_args, is_vararg) = self.compile_formal_arguments(formal_args)?;
+        let (formal_args, is_vararg) = self.compile_formal_arguments(formal_args_ast)?;
         let free_symbols = free_symbols(expr)?
             .iter()
             .inspect(|it| trace!("free: {}", it))
@@ -389,6 +432,7 @@ impl Vm {
             &free_symbols,
             is_vararg,
         );
+        lambda.set_desc(formal_args_ast.clone());
         if lambda.is_vararg {
             lambda.emit(OpCode::VarArg);
         }
