@@ -6,11 +6,13 @@ use crate::vm::gc::State;
 use crate::vm::lambda::Lambda;
 use crate::vm::vcell::VCell;
 use log::trace;
+use num::ToPrimitive;
 use std::collections::HashMap;
 use std::ops::Deref;
 
 #[derive(Debug)]
 pub struct Heap {
+    chunk_size: usize,
     free_list: Vec<usize>,
     heap: Vec<VCell>,
     heap_map: gc::Map,
@@ -24,6 +26,7 @@ impl Heap {
     /// chunk of free vcells.
     pub fn new(chunk_size: usize) -> Heap {
         Heap {
+            chunk_size,
             heap: vec![VCell::undefined(); chunk_size],
             free_list: (0..chunk_size).rev().into_iter().collect(),
             heap_map: gc::Map::new(chunk_size),
@@ -31,13 +34,36 @@ impl Heap {
         }
     }
 
+    /// Grow
+    ///
+    /// Grow the heap by one chunk, adding the newly created nodes
+    /// too the free list.
+    pub fn grow(&mut self) {
+        let current_size = self.heap.len();
+        let new_size = ((current_size / self.chunk_size) as f64 * 1.5)
+            .ceil()
+            .to_usize()
+            .unwrap()
+            * self.chunk_size;
+        self.heap.resize(new_size, VCell::undefined());
+        self.heap_map.resize(new_size);
+        (current_size..new_size).for_each(|it| self.free_list.push(it));
+    }
+
     /// Alloc
     ///
     /// Return the next free slot from the free list.
     pub fn alloc(&mut self) -> usize {
-        let ptr = self.free_list.pop().unwrap();
-        self.heap_map.set(ptr, State::Allocated);
-        ptr
+        match self.free_list.pop() {
+            None => {
+                self.grow();
+                self.alloc()
+            }
+            Some(ptr) => {
+                self.heap_map.set(ptr, State::Allocated);
+                ptr
+            }
+        }
     }
 
     /// Free
@@ -103,9 +129,8 @@ impl Heap {
             cell::Cell::String(ref s) => self.put(VCell::string(s.clone())),
             cell::Cell::Symbol(ref sym) => self.put(VCell::symbol(sym.clone())),
             cell::Cell::Continuation => panic!("unexpected continuation"),
-            cell::Cell::Closure => panic!("unexpected closure"),
             cell::Cell::Macro => panic!("unexpected macro"),
-            cell::Cell::Lambda => panic!("unexpected lambda"),
+            cell::Cell::Procedure(_) => panic!("unexpected lambda"),
             cell::Cell::Vector(ref vector) => {
                 let mut outv = Vec::with_capacity(vector.len());
                 for it in vector {
@@ -180,19 +205,36 @@ impl Heap {
             VCell::Char(val) => Cell::Char(*val),
             VCell::Number(val) => Cell::Number(val.clone()),
             VCell::Nil => Cell::Nil,
-            VCell::Pair(ref car, cdr) => Cell::new_pair(
-                self.get_as_cell(&VCell::Ptr(*car)),
-                self.get_as_cell(&VCell::Ptr(*cdr)),
-            ),
+            VCell::Pair(_, _) => {
+                let mut v = vec![];
+                let mut rest = vcell.clone();
+                loop {
+                    v.push(self.get_as_cell(&rest.as_car().unwrap()));
+                    match self.get_at_index(rest.as_cdr().unwrap().as_ptr().unwrap()) {
+                        pair if pair.is_pair() => {
+                            rest = pair.clone();
+                        }
+                        VCell::Nil => {
+                            return Cell::new_list(v);
+                        }
+                        cell => {
+                            return Cell::new_improper_list(v, self.get_as_cell(cell));
+                        }
+                    }
+                }
+            }
             VCell::Ptr(ptr) => self.get_as_cell(self.get_at_index(*ptr)),
             VCell::String(s) => Cell::String(s.borrow().deref().into()),
             VCell::Symbol(s) => Cell::Symbol(s.deref().into()),
             VCell::Undefined => Cell::Undefined,
             VCell::Void => Cell::Void,
             VCell::Continuation(_) => Cell::Continuation,
-            VCell::Closure(_, _) => Cell::Closure,
-            VCell::Lambda(_) => Cell::Lambda,
-            VCell::BuiltInProc(_) => Cell::Lambda,
+            VCell::Closure(ptr, _) => match self.get_at_index(*ptr).as_lambda() {
+                Ok(lambda) => Cell::Procedure(Some(lambda.to_string())),
+                Err(_) => Cell::Procedure(None),
+            },
+            VCell::Lambda(lambda) => Cell::Procedure(Some(lambda.to_string())),
+            VCell::BuiltInProc(proc) => Cell::Procedure(Some(proc.desc().to_string())),
             VCell::Macro(_) => Cell::Macro,
             VCell::Vector(vector) => {
                 let mut outv = Vec::with_capacity(vector.len());
@@ -424,6 +466,13 @@ impl Heap {
     pub fn used_size(&self) -> usize {
         self.capacity() - self.free_size()
     }
+
+    /// Chunk Size
+    ///
+    /// The chunk allocation size of this heap.
+    pub fn chunk_size(&self) -> usize {
+        self.chunk_size
+    }
 }
 
 #[cfg(test)]
@@ -532,4 +581,18 @@ mod tests {
         heap.sweep();
         assert_eq!(heap.free_list.len(), CHUNK_SIZE - 2);
     }
+
+    #[test]
+    fn heap_auto_grows() {
+        let mut heap = Heap::new(8);
+        assert_eq!(heap.chunk_size(), 8);
+        assert_eq!(heap.capacity(), 8);
+        assert_eq!(heap.free_size(), 8);
+        for _ in 0..9 {
+            heap.put_cell(&cell![0]);
+        }
+        assert_eq!(heap.chunk_size(), 8);
+        assert_eq!(heap.capacity(), 16);
+        assert_eq!(heap.free_size(), 7);
+    }
 }
