@@ -2,8 +2,8 @@ use crate::cell::Cell;
 use crate::vm::environment::{BindingSource, EnvironmentMap, LexicalEnvironment};
 use crate::vm::lambda::Lambda;
 use crate::vm::opcode::OpCode;
+use crate::vm::vcell::VCell;
 use crate::vm::vcell::VCell::LexicalEnvPtr;
-use crate::vm::vcell::{BuiltInProc, VCell};
 use crate::vm::Error::{
     InvalidBytecode, InvalidNumArgs, InvalidProcedure, InvalidSyntax, VariableNotBound,
 };
@@ -24,7 +24,7 @@ impl Vm {
         let mut cycles = 0;
         loop {
             cycles += 1;
-            if cycles % 256 == 0 {
+            if cycles % 8192 == 0 {
                 self.run_gc();
             }
             if cycles == count {
@@ -115,8 +115,9 @@ impl Vm {
                 let lambda = match self.heap.get(&self.acc) {
                     VCell::Closure(lambda, _) => lambda,
                     VCell::Lambda(_) => self.acc.as_ptr()?,
-                    VCell::BuiltInProc(BuiltInProc(func)) => {
-                        self.acc = match func(self)? {
+                    VCell::BuiltInProc(proc) => {
+                        let proc = proc.as_ref();
+                        self.acc = match proc.eval(self)? {
                             VCell::Ptr(ptr) => VCell::Ptr(ptr),
                             vcell => self.heap.put(vcell),
                         };
@@ -146,8 +147,9 @@ impl Vm {
                 let lambda = match self.heap.get(&self.acc) {
                     VCell::Closure(lambda, _) => lambda,
                     VCell::Lambda(_) => self.acc.as_ptr()?,
-                    VCell::BuiltInProc(BuiltInProc(func)) => {
-                        self.acc = match func(self)? {
+                    VCell::BuiltInProc(proc) => {
+                        let proc = proc.as_ref();
+                        self.acc = match proc.eval(self)? {
                             VCell::Ptr(ptr) => VCell::Ptr(ptr),
                             vcell => self.heap.put(vcell),
                         };
@@ -213,7 +215,7 @@ impl Vm {
                 let lambda = self.heap.get_at_index(lambda);
                 let lambda = lambda.as_lambda()?;
                 if self.stack.get_offset(-2)?.as_argc()? != lambda.args.len() {
-                    return Err(InvalidNumArgs("procedure".into()));
+                    return Err(InvalidNumArgs(lambda.to_string()));
                 }
 
                 self.stack.push(VCell::BasePointer(self.bp));
@@ -329,14 +331,16 @@ impl Vm {
     /// Read an argument vcell from program[ip], increment ip and
     /// return the value.
     fn read_operand(&mut self) -> Result<VCell, Error> {
-        let operand = self
-            .lambda()
-            .get(self.ip.1)
-            .cloned()
-            .filter(|it| !it.is_opcode())
-            .ok_or(InvalidBytecode);
-        self.ip.1 += 1;
-        operand
+        let proc = self.lambda();
+        match proc.get(self.ip.1) {
+            Some(opand) if opand.is_opcode() => Err(InvalidBytecode),
+            Some(opand) => {
+                let opand = opand.clone();
+                self.ip.1 += 1;
+                Ok(opand)
+            }
+            None => Err(InvalidBytecode),
+        }
     }
 
     /// Deref Arg
@@ -410,15 +414,16 @@ impl Vm {
     /// Read an op code from program[ip], increment ip and
     /// return the opcode.
     fn read_opcode(&mut self) -> Result<OpCode, Error> {
-        let op = self
-            .lambda()
-            .get(self.ip.1)
-            .cloned()
-            .filter(|it| it.is_opcode())
-            .map(|it| it.as_opcode())
-            .ok_or(InvalidBytecode)?;
-        self.ip.1 += 1;
-        op
+        let proc = self.lambda();
+        let op = proc.get(self.ip.1);
+        match op {
+            Some(op) => {
+                let op = op.as_opcode()?;
+                self.ip.1 += 1;
+                Ok(op)
+            }
+            None => Err(InvalidBytecode),
+        }
     }
 
     /// Load Arg
@@ -464,6 +469,11 @@ impl Vm {
         self.heap.mark(self.ip.0);
         self.heap.mark(self.ep);
         self.heap.sweep();
+
+        // If after GC the heap utilization is still high, grow the heap.
+        if (self.heap.used_size() as f64 / self.heap.capacity() as f64) > 0.75_f64 {
+            self.heap.grow();
+        }
     }
 
     /// Build Closure Environment
@@ -536,6 +546,17 @@ impl Vm {
         Ok(lexical_env)
     }
 
+    /// Pop
+    ///
+    /// Pop is a wrapper around vm.stack.pop(), providing automatic dereference
+    /// if the popped value is a heap reference.
+    pub fn pop(&mut self) -> Result<VCell, Error> {
+        match self.stack.pop()? {
+            VCell::Ptr(ptr) => Ok(self.heap.get_at_index(*ptr).clone()),
+            vcell => Ok(vcell.clone()),
+        }
+    }
+
     #[allow(dead_code)]
     fn trace_environment_map(&self, envmap: &EnvironmentMap) {
         trace!("--- ENVIRONMENT MAP ---");
@@ -554,6 +575,10 @@ impl Vm {
         trace!("--- END LEXICAL ENVIRONMENT ---");
     }
 
+    #[cfg(not(debug_assertions))]
+    fn trace_instruction(&self) {}
+
+    #[cfg(debug_assertions)]
     fn trace_instruction(&self) {
         trace!(
             "{:<60} {:>30}",
